#ifndef MYX_REDIS_PARSER_HPP_
#define MYX_REDIS_PARSER_HPP_

#pragma once

#include <myx/base/config.hpp>
#include <myx/redis/lexer.hpp>
#include <myx/redis/reply.hpp>

#include <QList>
#include <QPair>
#include <QVariant>

namespace myx {

namespace redis {

class Parser : public QObject
{
	Q_OBJECT

public:
	explicit Parser( Lexer*, QObject* = nullptr );

	Parser( const Parser& ) = delete;
	Parser& operator=( const Parser& ) = delete;
	Parser( Parser&& ) = delete;
	Parser& operator=( Parser&& ) = delete;

	~Parser() override = default;

	Q_SIGNAL void reply( const myx::redis::Reply& );
	Q_SIGNAL void replyMultiBulk( const myx::redis::Reply& );

private:
	Q_SLOT void readCharacter( char );
	Q_SLOT void readUnsafeString( const QString& );
	Q_SLOT void readSafeString( const QByteArray& );

	void descend();

	class Task
	{
		friend class Parser;

		enum { kUnknown = -2 };

		explicit Task( Reply::Type type ) :
			m_reply( type ),
			m_count( kUnknown ) {}

		Reply m_reply;
		int   m_count;
	};

	Lexer*        m_lexer;
	QList< Task > m_stack;

	Task& tos() { return( m_stack.last() ); }
}; // class Parser

} // namespace redis

} // namespace myx

#ifdef MYXLIB_HEADER_ONLY
#include "parser-inl.hpp"
#endif

#endif // ifndef MYX_REDIS_PARSER_HPP_
